'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _Type = require('./errors/Type');

var _Type2 = _interopRequireDefault(_Type);

var _chainType = require('./chainType');

var _chainType2 = _interopRequireDefault(_chainType);

var _isModel = require('./isModel');

var _isModel2 = _interopRequireDefault(_isModel);

var _isCollection = require('./isCollection');

var _isCollection2 = _interopRequireDefault(_isCollection);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Types
 */
var Types = {};

Types.string = (0, _chainType2.default)(function (value) {
  if (typeof value === 'undefined') {
    return value;
  }

  if (typeof value !== 'string') {
    throw new _Type2.default('value is not a string');
  }

  return value;
});

Types.bool = (0, _chainType2.default)(function (value) {
  if (typeof value === 'undefined') {
    return value;
  }

  if (typeof value !== 'boolean') {
    throw new _Type2.default('value is not a boolean');
  }

  return value;
});

Types.number = (0, _chainType2.default)(function (value) {
  if (typeof value === 'undefined') {
    return value;
  }

  if (typeof value !== 'number') {
    throw new _Type2.default('value is not a number');
  }

  return value;
});

Types.array = (0, _chainType2.default)(function (value) {
  if (typeof value === 'undefined') {
    return value;
  }

  if (!_lodash2.default.isArray(value)) {
    throw new _Type2.default('value is not an array');
  }

  return value;
});

Types.func = (0, _chainType2.default)(function (value) {
  if (typeof value === 'undefined') {
    return value;
  }

  if (typeof value !== 'function') {
    throw new _Type2.default('value is not a function');
  }

  return value;
});

Types.enum = function () {
  var enums = arguments.length <= 0 || arguments[0] === undefined ? [] : arguments[0];

  if (!_lodash2.default.isArray(enums)) {
    enums = [enums];
  }

  return (0, _chainType2.default)(function (value) {
    if (typeof value === 'undefined') {
      return value;
    }

    var isValid = enums.some(function (enumValue) {
      return value === enumValue;
    });

    if (isValid) {
      return value;
    }

    throw new _Type2.default('value is none of the provided enums');
  });
};

Types.enum.of = function () {
  var enumTypes = arguments.length <= 0 || arguments[0] === undefined ? [] : arguments[0];

  if (!_lodash2.default.isArray(enumTypes)) {
    enumTypes = [enumTypes];
  }

  return (0, _chainType2.default)(function (value) {
    if (typeof value === 'undefined') {
      return value;
    }

    var isValid = enumTypes.some(function (enumType) {
      try {
        enumType(value);

        return true;
      } catch (e) {
        return false;
      }
    });

    if (isValid) {
      return value;
    }

    throw new _Type2.default('value is none of the provided enum types');
  });
};

Types.any = (0, _chainType2.default)(function (value) {
  return value;
});

/**
 * Object
 */
Types.object = (0, _chainType2.default)(function (value) {
  if (typeof value === 'undefined') {
    return value;
  }

  if (!_lodash2.default.isPlainObject(value)) {
    throw new _Type2.default('value is not an object');
  }

  return value;
});

function validateAndReturnObject(value, schema) {
  return _lodash2.default.mapValues(schema, function (type, k) {
    try {
      return type(value[k]);
    } catch (e) {
      throw new _Type2.default('schema failed for key `' + k + '`, ' + e.message);
    }
  });
}

Types.object.of = function (schema) {
  if (!_lodash2.default.isPlainObject(schema)) {
    throw new _Type2.default('`object.of` must receive a plain object');
  }

  return (0, _chainType2.default)(function (value) {
    if (typeof value === 'undefined') {
      return value;
    }

    if (!_lodash2.default.isPlainObject(value)) {
      throw new _Type2.default('value is not an object');
    }

    return validateAndReturnObject(value, schema);
  });
};

/**
 * Model
 */
Types.model = (0, _chainType2.default)(function (value) {
  if (typeof value === 'undefined') {
    return value;
  }

  if ((0, _isModel2.default)(value)) {
    return value;
  }

  throw new _Type2.default('value is not a Model instance');
});

Types.model.of = function (Model) {
  if (typeof Model !== 'function') {
    throw new _Type2.default('Model is not a function');
  }

  return (0, _chainType2.default)(function (value) {
    if (typeof value === 'undefined') {
      return value;
    }

    if ((0, _isModel2.default)(value)) {
      if (value instanceof Model) {
        return value;
      }

      throw new _Type2.default('value is not instance of expected Model');
    }

    if (_lodash2.default.isPlainObject(value)) {
      return new Model(value);
    }

    throw new _Type2.default('value is not an object');
  });
};

/**
 * Collection
 */
Types.collection = (0, _chainType2.default)(function (value) {
  if (typeof value === 'undefined') {
    return value;
  }

  if ((0, _isCollection2.default)(value)) {
    return value;
  }

  throw new _Type2.default('value is not a Collection instance');
});

Types.collection.of = function (Collection) {
  if (typeof Collection !== 'function') {
    throw new _Type2.default('Collection is not a function');
  }

  return (0, _chainType2.default)(function (value) {
    if (typeof value === 'undefined') {
      return value;
    }

    if ((0, _isCollection2.default)(value)) {
      if (value instanceof Collection) {
        return value;
      }

      throw new _Type2.default('value is not instance of expected Collection');
    }

    if (_lodash2.default.isArray(value)) {
      return new Collection(value);
    }

    throw new _Type2.default('value is not an array');
  });
};

exports.default = Types;